# Store the version here so:
# 1) we don't load dependencies by storing it in __init__.py
# 2) hatchling can read it for dynamic versioning (pyproject.toml)
# 3) we can import it into your module
__version__ = "7.0.0"
